/*
 * $Id: mroute.c,v 1.2 1999/06/27 09:06:45 masaki Exp $
 */

#include <mrt.h>
#include <api6.h>
#include <igmp.h>
#include <net/route.h>
#ifdef HAVE_MULTICAST
/* symbols conflict */
#undef MRT_VERSION 
#ifdef linux
#include <linux/mroute.h>
#else
#include <netinet/ip_mroute.h>
#endif /* linux */

static int initialized = 0;
static int version = -1;

int
mc_mrtinit (void)
{
    int ret = -1;
    int yes = 1;

    if (initialized++)
	return (0);

    if ((ret = setsockopt (IGMP->sockfd, IPPROTO_IP, MRT_INIT, (char *) &yes,
		    sizeof (yes))) < 0) {
	trace (TR_ERROR, IGMP->trace, "setsockopt MRT_INIT (%d): %s\n",
	       yes, strerror (errno));
    }
    else {
	initialized++;
    }
    return (ret);
}


int
mc_mrtversion (void)
{
    int ret = -1;
    int verlen = sizeof (version);
   
    if (version > 0)
	return (version);
#ifdef MRT_VERSION
    if ((ret = getsockopt (IGMP->sockfd, IPPROTO_IP, MRT_VERSION, 
		           (char *) &version, &verlen)) < 0) {
	trace (TR_ERROR, IGMP->trace, "getsockopt MRT_VERSION: %s\n",
	       strerror (errno));
    }
#endif /* MRT_VERSION */
    if (ret >= 0)
	return (version);
    return (ret);
}


int
mc_mrtdone (void)
{
    int ret = -1;

    if (--initialized > 0)
	return (0);
    if ((ret = setsockopt (IGMP->sockfd, IPPROTO_IP, MRT_DONE, (char *) NULL,
		    0)) < 0) {
	trace (TR_ERROR, IGMP->trace, "setsockopt MRT_DONE: %s\n",
	       strerror (errno));
    }
    return (ret);
}


interface_t *vif2if[MAXVIFS];
int vif_refs[MAXVIFS];
int num_vif = 0;

/* not thread-safe but ok */
int
mc_add_vif (interface_t *interface)
{
    int vifi;
    struct vifctl vifc;
    int ret = -1;

    if (interface->vif_index >= 0) {
	assert (vif2if[interface->vif_index]);
	assert (vif_refs[interface->vif_index] > 0);
	vif_refs[interface->vif_index]++;
	return (0);
    }

    for (vifi = 0; vifi < num_vif; vifi++) {
	if (vif2if[vifi] == NULL)
	    break;
    }

    vifc.vifc_vifi = vifi;
    vifc.vifc_flags = 0; /* interface->mflags */
    if (interface->index < 0) {
        vifc.vifc_flags |= VIFF_TUNNEL;
        vifc.vifc_lcl_addr.s_addr = prefix_tolong (interface->tunnel_source);
        vifc.vifc_rmt_addr.s_addr = 
	    prefix_tolong (interface->tunnel_destination);
    }
    else {
        vifc.vifc_lcl_addr.s_addr = prefix_tolong (interface->primary->prefix);
    }
    vifc.vifc_threshold = 0; /* interface->threshold */
    vifc.vifc_rate_limit = 0; /* interface->rate_limit */

    /* this enables ALLMULTI on the interface to receive all multicast packets 
       so that it is required even for a normal interface  */
    if ((ret = setsockopt (IGMP->sockfd, IPPROTO_IP, MRT_ADD_VIF, 
			   (char *) &vifc, sizeof (vifc))) < 0) {
	trace (TR_ERROR, IGMP->trace, "setsockopt MRT_ADD_VIF (%d): %s\n",
	       vifi, strerror (errno));
    }
    else {
	assert (interface->vif_index < 0);
	interface->vif_index = vifi;
	assert (vif2if[vifi] == NULL);
	vif2if[vifi] = interface;
	assert (vif_refs[vifi] == 0);
	vif_refs[vifi]++;
	if (num_vif < ++vifi)
	    num_vif = vifi;
    }
    
    return (ret);
}


int
mc_del_vif (interface_t *interface)
{
    int vifi;
    struct vifctl vifc;
    int ret = -1;

    assert (interface->vif_index >= 0);
    assert (vif2if[interface->vif_index]);
    if (--vif_refs[interface->vif_index] > 0)
	return (0);

    vifc.vifc_vifi = interface->vif_index;
    vifc.vifc_flags = 0; /* interface->mflags */
    if (interface->index < 0) {
        vifc.vifc_flags |= VIFF_TUNNEL;
        vifc.vifc_lcl_addr.s_addr = prefix_tolong (interface->tunnel_source);
        vifc.vifc_rmt_addr.s_addr = 
	    prefix_tolong (interface->tunnel_destination);
    }
    else {
        vifc.vifc_lcl_addr.s_addr = prefix_tolong (interface->primary->prefix);
    }

    if ((ret = setsockopt (IGMP->sockfd, IPPROTO_IP, MRT_DEL_VIF, 
			   (char *) &vifc, sizeof (vifc))) < 0) {
	trace (TR_ERROR, IGMP->trace, "setsockopt MRT_DEL_VIF (%d): %s\n",
	       vifc.vifc_vifi, strerror (errno));
    }
    vif2if[vifc.vifc_vifi] = NULL;
    interface->vif_index = -1;
    for (vifi = num_vif; vifi > 0; vifi--)
        if (vif2if[vifi - 1] != NULL)
	    break;
    num_vif = vifi;
    
    return (ret);
}


int
mc_add_mfc (prefix_t *group, prefix_t *origin, interface_t *parent,
	    u_char *ttls)
{
    int ret = -1;
    struct mfcctl mfcctl;

    memset (&mfcctl, 0, sizeof (mfcctl));
    mfcctl.mfcc_mcastgrp.s_addr = prefix_tolong (group);
    mfcctl.mfcc_origin.s_addr = prefix_tolong (origin);
    mfcctl.mfcc_parent = (parent)?parent->vif_index:-1;
    if (ttls) {
	memcpy (mfcctl.mfcc_ttls, ttls, num_vif);
    }

    if ((ret = setsockopt (IGMP->sockfd, IPPROTO_IP, MRT_ADD_MFC, 
			  (char *) &mfcctl, sizeof (mfcctl))) < 0) {
	trace (TR_ERROR, IGMP->trace, 
	       "setsockopt MRT_ADD_MFC (group %s origin %s): %s\n",
	       prefix_toa (group), prefix_toa (origin), strerror (errno));
    }
    return (ret);
}


int
mc_del_mfc (prefix_t *group, prefix_t *origin)
{
    int ret = -1;
    struct mfcctl mfcctl;

    mfcctl.mfcc_mcastgrp.s_addr = prefix_tolong (group);
    mfcctl.mfcc_origin.s_addr = prefix_tolong (origin);

    if ((ret = setsockopt (IGMP->sockfd, IPPROTO_IP, MRT_DEL_MFC, 
			  (char *) &mfcctl, sizeof (mfcctl))) < 0) {
	trace (TR_ERROR, IGMP->trace, 
	       "setsockopt MRT_DEL_MFC (group %s origin %s): %s\n",
	       prefix_toa (group), prefix_toa (origin), strerror (errno));
    }
    return (ret);
}


int
mc_assert (int yes)
{
    int ret = -1;

#ifdef MRT_ASSERT
    if ((ret = setsockopt (IGMP->sockfd, IPPROTO_IP, MRT_ASSERT, 
			  (char *) &yes, sizeof (yes))) < 0) {
	trace (TR_ERROR, IGMP->trace, 
	       "setsockopt MRT_ASSERT (%d): %s\n", yes, strerror (errno));
    }
#endif /* MRT_ASSERT */
    return (ret);
}


#ifdef HAVE_IPV6
#ifndef SOLARIS_IPV6
#include <sys/sysctl.h>

int
mc6_mrtinit (void)
{
    int ret = -1;
    int one = 1;

    /* This lets all multicast-capable interfaces go to allmulti mode */
    /* there is no way to enable it per interface */
    if ((ret = sysctlbyname ("net.inet6.ipv6.mforwarding", 
		       NULL, NULL, &one, sizeof (one))) < 0) {
	trace (TR_ERROR, IGMPv6->trace, 
	       "sysctl (net.inet6.ipv6.mforwarding) = %d: %s\n", 
		one, strerror (errno));
    }
    return (ret);
}


int
mc6_mrtdone (void)
{
    int ret = -1;
    int zero = 0;

    if ((ret = sysctlbyname ("net.inet6.ipv6.mforwarding", 
		       NULL, NULL, &zero, sizeof (zero))) < 0) {
	trace (TR_ERROR, IGMPv6->trace, 
	       "sysctl (net.inet6.ipv6.mforwarding) = %d: %s\n", 
		zero, strerror (errno));
    }
    return (ret);
}
#endif /* SOLARIS_IPV6 */
#endif /* HAVE_IPV6 */

#endif /* HAVE_MULTICAST */
